"""Test Baseclass."""

from unittest import mock

import pytest

from pygeoif import geometry


def test_geometry_interface() -> None:
    """The geo interface must be implemented in subclasses."""
    base_geo = geometry._Geometry()

    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo.__geo_interface__


def test_bounds() -> None:
    """Subclasses must implement bounds."""
    base_geo = geometry._Geometry()
    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo.bounds


def test_wkt() -> None:
    """Implement wkt in subclasses."""
    base_geo = geometry._Geometry()
    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo.wkt


def test_empty() -> None:
    base_geo = geometry._Geometry()
    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo.is_empty


def test_wkt_inset() -> None:
    base_geo = geometry._Geometry()

    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo._wkt_inset == ""


def test_wkt_coordinates() -> None:
    base_geo = geometry._Geometry()
    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo._wkt_coords


def test_from_dict() -> None:
    base_geo = geometry._Geometry()
    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo._from_dict({"type": "_Geometry"})  # type: ignore


def test_has_z() -> None:
    base_geo = geometry._Geometry()
    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo.has_z


def test_convex_hull() -> None:
    with mock.patch("pygeoif.geometry._Geometry.has_z"):
        base_geo = geometry._Geometry()
        with pytest.raises(
            NotImplementedError,
            match="^Must be implemented by subclass$",
        ):
            assert base_geo.convex_hull


def test_get_bounds() -> None:
    base_geo = geometry._Geometry()
    with pytest.raises(NotImplementedError, match="^Must be implemented by subclass$"):
        assert base_geo._get_bounds()


@pytest.mark.parametrize(
    ("attr_val", "expected_error", "expected_error_message"),
    [
        # Happy path tests
        (
            ("attribute", "value"),
            AttributeError,
            "Attributes of _Geometry cannot be changed",
        ),
        (
            ("another_attribute", 123),
            AttributeError,
            "Attributes of _Geometry cannot be changed",
        ),
        (
            ("yet_another_attribute", [1, 2, 3]),
            AttributeError,
            "Attributes of _Geometry cannot be changed",
        ),
        # Edge cases
        (("", "value"), AttributeError, "Attributes of _Geometry cannot be changed"),
        ((None, "value"), TypeError, ".*attribute name must be string.*"),
        # Error cases
        ((123, "value"), TypeError, ".*attribute name must be string.*"),
        (([1, 2, 3], "value"), TypeError, ".*attribute name must be string.*"),
    ],
)
def test_setattr(attr_val, expected_error, expected_error_message) -> None:
    base_geo = geometry._Geometry()

    with pytest.raises(expected_error, match=f"^{expected_error_message}$"):
        setattr(base_geo, *attr_val)


@pytest.mark.parametrize(
    ("attr", "expected_error", "expected_error_message"),
    [
        (
            "attr1",
            AttributeError,
            "Attributes of _Geometry cannot be deleted",
        ),  # realistic test value
        (
            "",
            AttributeError,
            "Attributes of _Geometry cannot be deleted",
        ),  # edge case: empty string
        (
            None,
            TypeError,
            ".*attribute name must be string.*",
        ),  # edge case: None
        (
            123,
            TypeError,
            ".*attribute name must be string.*",
        ),  # error case: non-string attribute
    ],
    ids=[
        "realistic_test_value",
        "edge_case_empty_string",
        "edge_case_None",
        "error_case_non_string_attribute",
    ],
)
def test_delattr(attr, expected_error, expected_error_message) -> None:
    # Arrange
    base_geo = geometry._Geometry()

    # Act
    with pytest.raises(expected_error, match=f"^{expected_error_message}$"):
        delattr(base_geo, attr)



#
#   Copyright (C) 2012 -2022  Christian Ledermann
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.

#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.

#   You should have received a copy of the GNU Lesser General Public License
#   along with this library; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
"""Exceptions for pygeoif."""


class DimensionError(ValueError):
    """Geometries must have 2 or 3 dimensions."""


class WKTParserError(AttributeError):
    """WKT not supported or cannot be parsed."""


class InvalidGeometryError(ValueError):
    """Geometry is not valid."""


__all__ = [
    "DimensionError",
    "InvalidGeometryError",
    "WKTParserError",
]



"""Test LinearRing."""

from unittest import mock

import pytest

from pygeoif import exceptions
from pygeoif import functions
from pygeoif import geometry


def test_coords_get2d() -> None:
    ring = geometry.LinearRing([(0, 0), (1, 1), (2, 0)])

    assert ring.coords == ((0.0, 0.0), (1.0, 1.0), (2, 0), (0, 0))


def test_coords_get_3d() -> None:
    ring = geometry.LinearRing([(0, 0, 0), (1, 1, 1), (1, 2, 3)])

    assert ring.coords == ((0.0, 0.0, 0), (1.0, 1.0, 1), (1, 2, 3), (0, 0, 0))


def test_set_geoms_raises() -> None:
    ring = geometry.LinearRing([(0, 0), (1, 0)])  # pragma: no mutate

    with pytest.raises(
        exceptions.DimensionError,
        match="^All coordinates must have the same dimension$",
    ):
        ring._set_geoms([(0.0, 0.0, 0), (1.0, 1.0)])  # pragma: no mutate


def test_geo_interface() -> None:
    ring = geometry.LinearRing([(0, 0), (1, 1), (2, 2)])

    assert ring.__geo_interface__ == {
        "type": "LinearRing",
        "bbox": (0.0, 0.0, 2, 2),  # pragma: no mutate
        "coordinates": ((0.0, 0.0), (1.0, 1.0), (2, 2), (0, 0)),
    }


def test_bounds() -> None:
    ring = geometry.LinearRing([(1, 0), (3, 2)])

    assert ring.bounds == (1.0, 0.0, 3.0, 2.0)


def test_bounds3d() -> None:
    ring = geometry.LinearRing([(0, 0, 0), (1, 1, 3), (2, 2, 6)])  # pragma: no mutate

    assert ring.bounds == (0.0, 0.0, 2.0, 2.0)  # pragma: no mutate


def test_wkt() -> None:
    ring = geometry.LinearRing([(0, 0), (1, 1), (2, 2)])

    assert ring.wkt == "LINEARRING (0 0, 1 1, 2 2, 0 0)"


def test_wkt3d() -> None:
    ring = geometry.LinearRing([(0, 0, 0), (1, 1, 3), (2, 2, 6)])

    assert ring.wkt == "LINEARRING Z (0 0 0, 1 1 3, 2 2 6, 0 0 0)"


def test_from_dict() -> None:
    ring = geometry.LinearRing._from_dict(
        {
            "type": "LinearRing",
            "bbox": (0.0, 0.0, 1.0, 1.0),  # pragma: no mutate
            "coordinates": ((0.0, 0.0), (1.0, 1.0), (1, 2)),
        },
    )

    assert ring.coords == ((0.0, 0.0), (1.0, 1.0), (1, 2), (0, 0))


def test_from_compatible() -> None:
    not_a_geometry = mock.Mock(
        __geo_interface__={
            "type": "LinearRing",
            "coordinates": ((0.0, 0.0, 1.0), (1.0, 1.0, 2.0), (0, 4, 3)),
        },
    )

    ring = geometry.LinearRing._from_interface(not_a_geometry)

    assert isinstance(ring, geometry.LinearRing)
    assert ring.coords == ((0.0, 0.0, 1.0), (1.0, 1.0, 2.0), (0, 4, 3), (0, 0, 1))


def test_repr2d() -> None:
    ring = geometry.LinearRing([(0, 0), (1, 1), (2, 2)])

    assert repr(ring) == "LinearRing(((0, 0), (1, 1), (2, 2), (0, 0)))"


def test_repr3d() -> None:
    ring = geometry.LinearRing([(0, 0, 0), (1, 1, 3), (2, 2, 6)])

    assert repr(ring) == "LinearRing(((0, 0, 0), (1, 1, 3), (2, 2, 6), (0, 0, 0)))"


def test_repr_eval() -> None:
    ring = geometry.LinearRing([(0, 0, 0), (1, 1, 3), (2, 2, 6)])

    assert (
        eval(repr(ring), {}, {"LinearRing": geometry.LinearRing}).__geo_interface__
        == ring.__geo_interface__
    )


def test_signed_area() -> None:
    assert functions.signed_area(((0.0, 0.0), (1.0, 1.0), (2, 0), (0, 0))) == -1.0
    assert functions.signed_area(((0, 0, 5), (1, 0, 6), (1, 1, 7), (0, 0, 5))) == 0.5


def test_from_points() -> None:
    p1 = geometry.Point(0, 0)
    p2 = geometry.Point(1, 1)
    p3 = geometry.Point(0, 1)

    ring = geometry.LinearRing.from_points(p1, p2, p3)

    assert ring.coords == ((0, 0), (1, 1), (0, 1), (0, 0))


def test_convex_hull() -> None:
    line = geometry.LinearRing([(0, 0), (1, 1), (2, 2)])

    assert line.convex_hull == geometry.LineString([(0, 0), (2, 2)])


def test_convex_hull_3d() -> None:
    line = geometry.LinearRing([(0, 0, 0), (1, 1, 1), (2, 2, 2)])

    assert line.convex_hull == geometry.LineString([(0, 0), (2, 2)])


def test_convex_hull_3d_collapsed_to_point() -> None:
    line = geometry.LinearRing([(0, 0, 0), (0, 0, 1), (0, 0, 2)])

    assert line.convex_hull == geometry.Point(0, 0)


def test_convex_hull_linear_ring() -> None:
    line = geometry.LinearRing([(0, 0), (1, 0), (2, 2)])

    assert line.convex_hull == geometry.Polygon([(0, 0), (1, 0), (2, 2), (0, 0)])


def test_is_ccw() -> None:
    line = geometry.LinearRing([(0, 0), (1, 0), (1, 1), (0, 0)])

    assert line.is_ccw


def test_is_cw() -> None:
    line = geometry.LinearRing([(0, 0), (1, 1), (1, 0), (0, 0)])

    assert not line.is_ccw


def test_centroid_line() -> None:
    line = geometry.LinearRing([(0, 0), (0, 1)])

    assert line.centroid is None


def test_centroid_3d() -> None:
    line = geometry.LinearRing([(0, 0, 1), (2, 0, 2), (2, 2, 0), (0, 2, 0)])

    with pytest.raises(
        exceptions.DimensionError,
        match="^Centeroid is only implemented for 2D coordinates$",
    ):
        assert line.centroid


def test_centroid_crossing() -> None:
    line = geometry.LinearRing([(0, 0), (1, 0), (1, 1), (0, -1)])

    assert line.centroid is None


def test_centroid_valid() -> None:
    line = geometry.LinearRing([(0, 0), (4, 0), (4, 2), (0, 2)])

    assert line.centroid == geometry.Point(2, 1)


def test_empty() -> None:
    ring = geometry.LinearRing([])

    assert ring.is_empty


def test_empty_bounds() -> None:
    ring = geometry.LinearRing([])

    assert ring.bounds == ()



#
#   Copyright (C) 2012 - 2024 Christian Ledermann
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.

#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.

#   You should have received a copy of the GNU Lesser General Public License
#   along with this library; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
"""Types for geometries."""
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from typing_extensions import Literal
from typing_extensions import NotRequired
from typing_extensions import Protocol
from typing_extensions import TypedDict  # for Python <3.11 with (Not)Required

Point2D = Tuple[float, float]
Point3D = Tuple[float, float, float]
PointType = Union[Point2D, Point3D]
Line2D = Sequence[Point2D]
Line3D = Sequence[Point3D]
LineType = Union[Line2D, Line3D]
Interiors = Optional[Sequence[LineType]]
Poly2d = Union[
    Tuple[Line2D, Sequence[Line2D]],
    Tuple[Line2D],
]
Poly3d = Union[
    Tuple[Line3D, Sequence[Line3D]],
    Tuple[Line3D],
]
PolygonType = Union[
    Poly2d,
    Poly3d,
]

MultiGeometryType = Sequence[Union[PointType, LineType, PolygonType]]
Bounds = Tuple[float, float, float, float]

CoordinatesType = Union[
    PointType,
    LineType,
    Sequence[LineType],
]
MultiCoordinatesType = Sequence[CoordinatesType]

GeomType = Literal[
    "Point",
    "LineString",
    "LinearRing",
    "Polygon",
    "MultiPoint",
    "MultiLineString",
    "MultiPolygon",
]


class GeoInterface(TypedDict):
    """Required keys for the GeoInterface."""

    type: GeomType
    coordinates: Union[CoordinatesType, MultiCoordinatesType]
    bbox: NotRequired[Bounds]


class GeoCollectionInterface(TypedDict):
    """Geometry Collection Interface."""

    type: Literal["GeometryCollection"]
    geometries: Sequence[Union[GeoInterface, "GeoCollectionInterface"]]
    bbox: NotRequired[Bounds]


class GeoFeatureInterface(TypedDict):
    """The GeoFeatureInterface has optional keys."""

    type: Literal["Feature"]
    bbox: NotRequired[Bounds]
    properties: NotRequired[Dict[str, Any]]
    id: NotRequired[Union[str, int]]
    geometry: GeoInterface


class GeoFeatureCollectionInterface(TypedDict):
    """Bbox and id are optional keys for the GeoFeatureCollectionInterface."""

    type: Literal["FeatureCollection"]
    features: Sequence[GeoFeatureInterface]
    bbox: NotRequired[Bounds]
    id: NotRequired[Union[str, int]]


class GeoType(Protocol):
    """Any compatible type that implements the __geo_interface__."""

    @property
    def __geo_interface__(self) -> GeoInterface:
        """Return the GeoInterface."""


class GeoCollectionType(Protocol):
    """Any compatible type that implements the __geo_interface__."""

    @property
    def __geo_interface__(self) -> GeoCollectionInterface:
        """Return the GeoInterface."""


__all__ = [
    "Bounds",
    "CoordinatesType",
    "Interiors",
    "GeoCollectionInterface",
    "GeoCollectionType",
    "GeoFeatureCollectionInterface",
    "GeoFeatureInterface",
    "GeoInterface",
    "GeoType",
    "LineType",
    "MultiCoordinatesType",
    "MultiGeometryType",
    "Point2D",
    "Point3D",
    "PointType",
    "PolygonType",
]



#
#   Copyright (C) 2012 -2022  Christian Ledermann
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.

#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.

#   You should have received a copy of the GNU Lesser General Public License
#   along with this library; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# file deepcode ignore inconsistent~equality: Python 3 only
"""Features."""
from typing import Any
from typing import Dict
from typing import Generator
from typing import Iterator
from typing import Optional
from typing import Sequence
from typing import Union
from typing import cast

from pygeoif.functions import compare_coordinates
from pygeoif.geometry import Geometry
from pygeoif.types import Bounds
from pygeoif.types import GeoFeatureCollectionInterface
from pygeoif.types import GeoFeatureInterface


def feature_geo_interface_equals(
    my_interface: GeoFeatureInterface,
    other_interface: GeoFeatureInterface,
) -> bool:
    """Check if my interface is the same as the other interface."""
    return all(
        [
            my_interface.get("id") == other_interface.get("id"),
            my_interface["type"] == other_interface.get("type"),
            my_interface["properties"] == other_interface.get("properties"),
            my_interface["geometry"]["type"] == other_interface["geometry"].get("type"),
            compare_coordinates(
                coords=my_interface["geometry"]["coordinates"],
                other=other_interface["geometry"].get(  # type: ignore [arg-type]
                    "coordinates",
                ),
            ),
        ],
    )


class Feature:
    """
    Aggregates a geometry instance with associated user-defined properties.

    Attributes:
    ----------
    geometry : object
        A geometry instance
    properties : dict
        A dictionary linking field keys with values
        associated with geometry instance

    Example:
    -------
     >>> p = Point(1.0, -1.0)
     >>> props = {'Name': 'Sample Point', 'Other': 'Other Data'}
     >>> a = Feature(p, props)
     >>> a.properties
     {'Name': 'Sample Point', 'Other': 'Other Data'}
      >>> a.properties['Name']
     'Sample Point'

    """

    def __init__(
        self,
        geometry: Geometry,
        properties: Optional[Dict[str, Any]] = None,
        feature_id: Optional[Union[str, int]] = None,
    ) -> None:
        """Initialize the feature."""
        self._geometry = geometry
        self._properties = properties or {}
        self._feature_id = feature_id

    def __eq__(self, other: object) -> bool:
        """Check if the geointerfaces are equal."""
        try:
            if not other.__geo_interface__.get(  # type: ignore [attr-defined]
                "geometry",
            ):
                return False
        except AttributeError:
            return False
        return feature_geo_interface_equals(
            my_interface=self.__geo_interface__,
            other_interface=other.__geo_interface__,  # type: ignore [attr-defined]
        )

    def __repr__(self) -> str:
        """Return the representation."""
        return (
            f"{self.__class__.__name__}({self._geometry!r},"
            f" {self._properties}, {self._feature_id!r})"
        )

    @property
    def id(self) -> Optional[Union[str, int]]:
        """Return the id of the feature."""
        return self._feature_id

    @property
    def geometry(self) -> Geometry:
        """Return the geometry of the feature."""
        return self._geometry

    @property
    def properties(self) -> Dict[str, Any]:
        """Return a dictionary of properties."""
        return self._properties

    @property
    def __geo_interface__(self) -> GeoFeatureInterface:
        """Return the GeoInterface of the geometry with properties."""
        geo_interface: GeoFeatureInterface = {
            "type": "Feature",
            "bbox": cast(Bounds, self._geometry.bounds),
            "geometry": self._geometry.__geo_interface__,
            "properties": self._properties,
        }
        if self._feature_id is not None:
            geo_interface["id"] = self._feature_id

        return geo_interface


class FeatureCollection:
    """
    A heterogenous collection of Features.

    Attributes:
    ----------
    features : sequence
        A sequence of feature instances


    Example:
    -------
    >>> from pygeoif import geometry
    >>> p = geometry.Point(1.0, -1.0)
    >>> props = {'Name': 'Sample Point', 'Other': 'Other Data'}
    >>> a = geometry.Feature(p, props)
    >>> p2 = geometry.Point(1.0, -1.0)
    >>> props2 = {'Name': 'Sample Point2', 'Other': 'Other Data2'}
    >>> b = geometry.Feature(p2, props2)
    >>> features = [a, b]
    >>> c = geometry.FeatureCollection(features)
    >>> c.__geo_interface__
    {'type': 'FeatureCollection',
     'features': [{'geometry': {'type': 'Point', 'coordinates': (1.0, -1.0)},
     'type': 'Feature',
     'properties': {'Other': 'Other Data', 'Name': 'Sample Point'}},
    {'geometry': {'type': 'Point', 'coordinates': (1.0, -1.0)},
     'type': 'Feature',
     'properties': {'Other': 'Other Data2', 'Name': 'Sample Point2'}}]}

    """

    def __init__(self, features: Sequence[Feature]) -> None:
        """Initialize the feature."""
        self._features = tuple(features)

    def __eq__(self, other: object) -> bool:
        """Check if the geointerfaces are equal."""
        return self._check_interface(other) and all(
            (
                feature_geo_interface_equals(my_interface=mine, other_interface=other)
                for mine, other in zip(
                    self.__geo_interface__["features"],
                    other.__geo_interface__["features"],  # type: ignore [attr-defined]
                )
            ),
        )

    def __len__(self) -> int:
        """Return the umber of features in this collection."""
        return len(self._features)

    def __iter__(self) -> Iterator[Feature]:
        """Iterate over the features of the collection."""
        return iter(self._features)

    def __repr__(self) -> str:
        """Return the representation."""
        return f"{self.__class__.__name__}({self._features!r})"

    @property
    def features(self) -> Generator[Feature, None, None]:
        """Iterate over the features of the collection."""
        yield from self._features

    @property
    def bounds(self) -> Bounds:
        """Return the X-Y bounding box."""
        geom_bounds = list(
            zip(*(feature.geometry.bounds for feature in self._features)),
        )
        return (
            min(geom_bounds[0]),
            min(geom_bounds[1]),
            max(geom_bounds[2]),
            max(geom_bounds[3]),
        )

    @property
    def __geo_interface__(self) -> GeoFeatureCollectionInterface:
        """Return the GeoInterface of the feature."""
        return {
            "type": "FeatureCollection",
            "bbox": self.bounds,
            "features": tuple(feature.__geo_interface__ for feature in self._features),
        }

    def _check_interface(self, other: object) -> bool:
        try:
            return self.__geo_interface__[
                "type"
            ] == other.__geo_interface__.get(  # type: ignore [attr-defined]
                "type",
            ) and len(
                self.__geo_interface__["features"],
            ) == len(
                other.__geo_interface__.get(  # type: ignore [attr-defined]
                    "features",
                    [],
                ),
            )
        except AttributeError:
            return False


__all__ = [
    "Feature",
    "FeatureCollection",
]



"""Test MultiPoint."""

import math

import pytest

from pygeoif import geometry


def test_geoms() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 1), (1, 2), (2, 2)])

    for point in multipoint.geoms:
        assert type(point) is geometry.Point


def test_len() -> None:
    multipoint = geometry.MultiPoint(
        [(0, 0), (1, 1), (1, 2), (2, 2), (0, 0), (1, 1), (1, 2), (2, 2)],
    )

    assert len(multipoint) == 8


def test_bounds() -> None:
    multipoint = geometry.MultiPoint(
        [(0, 1), (1, 1), (3, 2)],
    )

    assert multipoint.bounds == (0, 1, 3, 2)


def test_has_z_empty() -> None:
    multipoint = geometry.MultiPoint(())

    assert multipoint.has_z is None


def test_geo_interface() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 1), (1, 2), (2, 2)])

    assert multipoint.__geo_interface__ == {
        "type": "MultiPoint",
        "bbox": (0, 0, 2, 2),
        "coordinates": ((0, 0), (1, 1), (1, 2), (2, 2)),
    }


def test_from_dict() -> None:
    multipoint = geometry.MultiPoint._from_dict(
        {
            "type": "MultiPoint",
            "bbox": (0.0, 0.0, 1.0, 1.0),  # pragma: no mutate
            "coordinates": ((0.0, 0.0), (1.0, 1.0)),
        },
    )

    assert multipoint.__geo_interface__ == {
        "type": "MultiPoint",
        "bbox": (0.0, 0.0, 1.0, 1.0),  # pragma: no mutate
        "coordinates": ((0.0, 0.0), (1.0, 1.0)),
    }


def test_coords() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 1), (1, 2), (2, 2)])

    with pytest.raises(
        NotImplementedError,
        match="^Multi-part geometries do not provide a coordinate sequence$",
    ):
        assert multipoint.coords


def test_unique() -> None:
    multipoint = geometry.MultiPoint(
        [(0, 0), (1, 1), (1, 2), (2.0, 2.0), (0, 0), (1.0, 1.0), (1, 2), (2, 2)],
        unique=True,
    )

    assert len(multipoint) == 4


def test_wkt() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 1), (1, 2), (2, 2)])

    assert multipoint.wkt == "MULTIPOINT (0 0, 1 1, 1 2, 2 2)"


def test_repr() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 1), (1, 2), (2, 2)])

    assert repr(multipoint) == "MultiPoint(((0, 0), (1, 1), (1, 2), (2, 2)))"


def test_repr_eval() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 1), (1, 2), (2, 2)])

    assert (
        eval(
            repr(multipoint),
            {},
            {"MultiPoint": geometry.MultiPoint},
        ).__geo_interface__
        == multipoint.__geo_interface__
    )


def test_convex_hull() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 1), (2, 2)])

    assert multipoint.convex_hull == geometry.LineString([(0, 0), (2, 2)])


def test_convex_hull_3d() -> None:
    multipoint = geometry.MultiPoint([(0, 0, 0), (1, 1, 1), (2, 2, 2)])

    assert multipoint.convex_hull == geometry.LineString([(0, 0), (2, 2)])


def test_convex_hull_3d_collapsed_to_point() -> None:
    multipoint = geometry.MultiPoint([(0, 0, 0), (0, 0, 1), (0, 0, 2)])

    assert multipoint.convex_hull == geometry.Point(0, 0)


def test_convex_hull_linear_ring() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 0), (2, 2)])

    assert multipoint.convex_hull == geometry.Polygon([(0, 0), (1, 0), (2, 2), (0, 0)])


def test_from_points() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 0), (2, 2)])
    p1 = geometry.Point(0, 0)
    p2 = geometry.Point(1, 0)
    p3 = geometry.Point(2.0, 2.0)

    assert geometry.MultiPoint.from_points(p1, p2, p3) == multipoint


def test_from_points_unique() -> None:
    multipoint = geometry.MultiPoint([(0, 0), (1, 0), (2, 2)], unique=True)
    p1 = geometry.Point(0, 0)
    p2 = geometry.Point(1, 0)
    p3 = geometry.Point(2.0, 2.0)

    assert (
        geometry.MultiPoint.from_points(p1, p2, p3, p1, p2, p3, p1, unique=True)
        == multipoint
    )


def test_empty() -> None:
    multipoint = geometry.MultiPoint([(1, None)])

    assert multipoint.is_empty


def test_repr_empty() -> None:
    multipoint = geometry.MultiPoint([(math.nan, math.nan)])

    assert repr(multipoint) == "MultiPoint(())"


def test_empty_bounds() -> None:
    multipoint = geometry.MultiPoint([(None, None)])

    assert multipoint.bounds == ()


def test_empty_geoms() -> None:
    multipoint = geometry.MultiPoint([(math.nan, math.nan)])

    assert not list(multipoint.geoms)



"""Optimize bounds for multigeometries."""

minx = (1, 2, 3, 4)
miny = (8, 7, 6, 5)
maxx = (5, 6, 7, 8)
maxy = (9, 8, 7, 6)


def get_bounds():
    return zip(minx, miny, maxx, maxy)


def test_bounds() -> None:
    assert list(get_bounds()) == [
        (1, 8, 5, 9),
        (2, 7, 6, 8),
        (3, 6, 7, 7),
        (4, 5, 8, 6),
    ]


def test_brute() -> None:
    new_bounds = (
        min(b[0] for b in get_bounds()),
        min(b[1] for b in get_bounds()),
        max(b[2] for b in get_bounds()),
        max(b[3] for b in get_bounds()),
    )

    assert new_bounds == (1, 5, 8, 9)


def test_unzip() -> None:
    assert list(zip(*get_bounds())) == [minx, miny, maxx, maxy]


def test_bounds_unzipped() -> None:
    tb = list(zip(*get_bounds()))
    new_bounds = (
        min(tb[0]),
        min(tb[1]),
        max(tb[2]),
        max(tb[3]),
    )

    assert new_bounds == (1, 5, 8, 9)


def get_line():
    return zip(minx, miny)


def test_line() -> None:
    assert list(get_line()) == [
        (1, 8),
        (2, 7),
        (3, 6),
        (4, 5),
    ]


def test_unzip_line() -> None:
    assert list(zip(*get_line())) == [(1, 2, 3, 4), (8, 7, 6, 5)]


def test_line_bounds() -> None:
    xy = list(zip(*get_line()))

    bounds = (
        min(xy[0]),
        min(xy[1]),
        max(xy[0]),
        max(xy[1]),
    )

    assert bounds == (1, 5, 4, 8)



"""Test Feature and FeatureCollection."""

import unittest

import pytest

from pygeoif import feature
from pygeoif import geometry


class TestFeature:
    def setup_method(self) -> None:
        self.a = geometry.Polygon(
            ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)),
        )
        self.b = geometry.Polygon(
            ((0.0, 0.0), (0.0, 2.0), (2.0, 1.0), (2.0, 0.0), (0.0, 0.0)),
        )
        self.f1 = feature.Feature(self.a)
        self.f2 = feature.Feature(self.b)
        self.f3 = feature.Feature(self.a, {}, feature_id="1")
        self.fc = feature.FeatureCollection([self.f1, self.f2])

    def test_feature_eq(self) -> None:
        assert self.f1 == feature.Feature(self.a)
        assert self.f3 == feature.Feature(self.a, {}, feature_id="1")

    def test_feature_neq_no_geo_interface(self) -> None:
        assert self.f1 != object()

    def test_feature_neq_no_geo_interface_geometry(self) -> None:
        assert self.f1 != unittest.mock.Mock(__geo_interface__={})

    def test_feature_neq_no_geo_interface_type(self) -> None:
        assert self.f1 != unittest.mock.Mock(__geo_interface__={"type": "foo"})

    def test_feature_neq_no_geo_interface_features(self) -> None:
        assert self.f1 != unittest.mock.Mock(__geo_interface__={"type": "Feature"})

    def test_feature(self) -> None:
        pytest.raises(TypeError, feature.Feature)
        assert self.f1.__geo_interface__ == {
            "type": "Feature",
            "bbox": (0.0, 0.0, 1.0, 1.0),
            "geometry": {
                "type": "Polygon",
                "bbox": (0.0, 0.0, 1.0, 1.0),
                "coordinates": (
                    ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)),
                ),
            },
            "properties": {},
        }
        self.f1.properties["coords"] = {"cube": (0, 0, 0)}
        assert self.f1.__geo_interface__ == {
            "type": "Feature",
            "bbox": (0.0, 0.0, 1.0, 1.0),
            "geometry": {
                "type": "Polygon",
                "bbox": (0.0, 0.0, 1.0, 1.0),
                "coordinates": (
                    ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)),
                ),
            },
            "properties": {"coords": {"cube": (0, 0, 0)}},
        }
        assert self.f1.geometry.bounds == (0.0, 0.0, 1.0, 1.0)
        del self.f1.properties["coords"]

    def test_feature_with_id(self) -> None:
        assert self.f3.id == "1"
        assert self.f3.__geo_interface__ == {
            "type": "Feature",
            "bbox": (0.0, 0.0, 1.0, 1.0),
            "geometry": {
                "type": "Polygon",
                "bbox": (0.0, 0.0, 1.0, 1.0),
                "coordinates": (
                    ((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)),
                ),
            },
            "id": "1",
            "properties": {},
        }

    def test_feature_repr(self) -> None:
        assert (
            repr(self.f3) == "Feature("
            "Polygon(((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)),),"
            " {}, '1')"
        )

    def test_feature_repr_eval(self) -> None:
        assert (
            eval(
                repr(self.f2),
                {},
                {"Polygon": geometry.Polygon, "Feature": feature.Feature},
            ).__geo_interface__
            == self.f2.__geo_interface__
        )

    def test_featurecollection(self) -> None:
        pytest.raises(TypeError, feature.FeatureCollection)
        pytest.raises(TypeError, feature.FeatureCollection, None)
        assert len(list(self.fc.features)) == 2
        assert len(self.fc) == 2
        assert self.fc.bounds == (0.0, 0.0, 2.0, 2.0)
        assert [self.f1, self.f2] == list(self.fc)
        assert self.fc.__geo_interface__ == {
            "bbox": (0.0, 0.0, 2.0, 2.0),
            "features": (
                {
                    "bbox": (0.0, 0.0, 1.0, 1.0),
                    "geometry": {
                        "bbox": (0.0, 0.0, 1.0, 1.0),
                        "coordinates": (
                            (
                                (0.0, 0.0),
                                (0.0, 1.0),
                                (1.0, 1.0),
                                (1.0, 0.0),
                                (0.0, 0.0),
                            ),
                        ),
                        "type": "Polygon",
                    },
                    "properties": {},
                    "type": "Feature",
                },
                {
                    "bbox": (0.0, 0.0, 2.0, 2.0),
                    "geometry": {
                        "bbox": (0.0, 0.0, 2.0, 2.0),
                        "coordinates": (
                            (
                                (0.0, 0.0),
                                (0.0, 2.0),
                                (2.0, 1.0),
                                (2.0, 0.0),
                                (0.0, 0.0),
                            ),
                        ),
                        "type": "Polygon",
                    },
                    "properties": {},
                    "type": "Feature",
                },
            ),
            "type": "FeatureCollection",
        }

    def test_featurecollection_eq(self) -> None:
        assert self.fc == feature.FeatureCollection([self.f1, self.f2])

    def test_featurecollection_neq_no_geo_interface(self) -> None:
        assert self.fc != object()

    def test_featurecollection_neq_no_geo_interface_geometry(self) -> None:
        assert self.fc != unittest.mock.Mock(__geo_interface__={})

    def test_featurecollection_neq_no_geo_interface_features(self) -> None:
        assert self.fc != unittest.mock.Mock(
            __geo_interface__={"type": "FeatureCollection"},
        )

    def test_featurecollection_neq_no_geo_interface_len_features(self) -> None:
        assert self.fc != feature.FeatureCollection([self.f1])

    def test_featurecollection_repr(self) -> None:
        assert (
            repr(self.fc) == "FeatureCollection("
            "(Feature("
            "Polygon(((0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)),),"
            " {}, None), "
            "Feature("
            "Polygon(((0.0, 0.0), (0.0, 2.0), (2.0, 1.0), (2.0, 0.0), (0.0, 0.0)),), "
            "{}, None)))"
        )

    def test_featurecollection_repr_eval(self) -> None:
        assert (
            eval(
                repr(self.fc),
                {},
                {
                    "Polygon": geometry.Polygon,
                    "Feature": feature.Feature,
                    "FeatureCollection": feature.FeatureCollection,
                },
            ).__geo_interface__
            == self.fc.__geo_interface__
        )

    def test_featurecollection_bounds(self) -> None:
        ls1 = geometry.LineString(((0, 1), (1, 1)))
        ls2 = geometry.LineString(((2, 3), (3, 4)))
        fc = feature.FeatureCollection([feature.Feature(ls1), feature.Feature(ls2)])

        assert fc.bounds == (0, 1, 3, 4)
